/*-
 * Copyright (C)2005..2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2005..2021 @BABOLO http://www.babolo.ru/\n"
#ident "@(#) $Id: popushd.c,v 1.24 2021/09/01 00:36:55 babolo Exp $\n"

#define BLIN_COMPAT  4
#define MIFE_COMPAT  5
#define Bpars_COMPAT 4
#define MULAR_COMPAT 0

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <syslog.h>
#include <stdarg.h>
#include <syslog.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <multilar.h>
#include <mife.h>
#include "popushd.h"

u_int32_t popush_state;

static void
sighup(int sig) {
    popush_state |= POPUSH_FSIGN;
}

static void
sigint(int sig) {
    popush_state |= POPUSH_FSTOP;
}

static void
usage(int ex) {
    fprintf( stderr
           , "popushd @BABOLO V.M "VERS"  "DATE" http://www.babolo.ru/\n"
             "Usage:\n"
             "popushd [-qv ][-f conf][-p pid]\n"
             "    -. envname  - set envname for cwd\n"
             "    -@ envname  - set envname for PID\n"
             "    -a 0xXX     - absolute quote\n"
             "    -d          - not daemonize\n"
             "    -f file     - config file\n"
             "    -p file     - PID file\n"
             "    -q          - quiet\n"
             "    -v[v[v...]] - verbose\n"
             "    -V          - dump and exit\n"
             "    -?          - this text\n"
             )
    ;
    exit(ex);
}

int
main(int argc, char *argv[], char **envp) {
    const char   *flar    = ".:@:a:df:p:qvV=:?";
    int           ex      = EX_OK;
    const char   *emyspid = NULL;
    babolo_opts  *bos     = NULL;
    char          myspid[8];
    FILE         *pidfile;
    struct kevent evnt;
    pid_t         mypid;
    popush_main  *main;
    popush_cnf   *cfg;
    const char   *o;
    int           c;

    if  (!(main = calloc(1, sizeof(popush_main)))) {
        ifBLIN_QW0("No mem");
        ex = -EX_OSERR;
        goto out;
    }
    blin_ctl(BLIN_CTL_FLAG | BLIN_CTL_FSET, (main->flags = BLIN_BIT0));
#   define blin_internal_flags (main->flags & BLIN_MASK)
    main->cnfname = RCFILE;
    main->pidname = PIDFILE;
    if  (!(main->wdpath = getcwd(NULL, 0))) {
        ifBLIN_QW1("getcwd failed");
    }
    if  (!(bos = babolo_openopts(Bpars_NONU | Bpars_SYME, 'Z'))) {
        ifBLIN_QW0("babolo_openopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (babolo_setopts(bos, Bpars_NONU, argc, argv, flar)) {
        ifBLIN_QW0("babolo_setopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    while (0 < (c = babolo_getopts(bos))) {
        if  (0 > c) {
            ifBLIN_QW0("babolo_getopts");
            ex = EX_USAGE;
            goto out;
        }
        switch (c) {
        case '.':
            if  (!(o = babolo_getoptsarg(bos))) {
                ifBLIN_QW0("No -%c arg", bos->c);
                ex = EX_USAGE;
                goto out;
            }
            if  (setenv(o, main->wdpath, 1) < 0) {
                ifBLIN_QW0("setenv %s=%s", o, main->wdpath);
            }
            break;
        case '@':
            if  (!(o = babolo_getoptsarg(bos))) {
                ifBLIN_QW0("No -%c arg", bos->c);
                ex = EX_USAGE;
                goto out;
            }
            if  (!(emyspid = strdup(o))) {
                ifBLIN_QW0("strdup -%c arg", bos->c);
                ex = EX_OSERR;
                goto out;
            }
            break;
        case 'a':
            if  (!(o = babolo_getoptsarg(bos))) {
                ifBLIN_QW0("No -%c arg", bos->c);
                ex = EX_USAGE;
                goto out;
            }
            main->absq = strtol(o, NULL, 0);
            break;
        case 'd':
            main->flags |= POPUSH_DMN;
            break;
        case 'f':
            if  (!(o = babolo_getoptsarg(bos))) {
                ifBLIN_QW0("No -%c arg", bos->c);
                ex = EX_USAGE;
                goto out;
            }
            if  (!(main->cnfname = strdup(o))) {
                ifBLIN_QW0("strdup -%c arg", bos->c);
                ex = EX_OSERR;
                goto out;
            }
            break;
        case 'p':
            if  (!(o = babolo_getoptsarg(bos))) {
                ifBLIN_QW0("No -%c arg", bos->c);
                ex = EX_USAGE;
                goto out;
            }
            if  (!(main->pidname = strdup(o))) {
                ifBLIN_QW0("strdup -%c arg", bos->c);
                ex = EX_OSERR;
                goto out;
            }
            break;
        case 'q':
            BLIN_QUIET(main->flags);
            break;
        case 'v':
            BLIN_VERBOSE(main->flags);
            break;
        case 'V':
            main->flags |= POPUSH_ANA;
            break;
        case '?':
            usage(EX_OK);
        case '=':
            if  (!(o = babolo_getoptsarg(bos))) {
                ifBLIN_QW0("No -%c arg", bos->c);
                ex = EX_USAGE;
                goto out;
            }
            switch(o[0]) {
            case '0': blin_ctl(BLIN_CTL_DUMP); exit(EX_OK);
            case '1': babolo_optest(flar);     exit(EX_OK);
            case '2': babolo_optext(bos);      exit(EX_OK);
            case '3': babolo_dumpopts(bos);    exit(EX_OK);
            }
            break;
        default:
            usage(EX_USAGE);
    }   }
    blin_ctl( BLIN_BIT0 | BLIN_BIT1 | BLIN_CTL_WHAT
            , ((main->flags & POPUSH_DMN) ? BLIN_MODWARN : BLIN_MODLOGR) | BLIN_MODEOAU
            );
    blin_ctl(BLIN_BIT0 | BLIN_CTL_PRIO, LOG_DAEMON | LOG_CRIT);
    blin_ctl(BLIN_BIT1 | BLIN_CTL_PRIO, LOG_DAEMON | LOG_ERR);
    blin_ctl(BLIN_BIT2 | BLIN_CTL_PRIO, LOG_DAEMON | LOG_WARNING);
    if  (main->wdpath && main->cnfname && (main->cnfname[0] !=  '/')) {
        char *t;

        if  (asprintf(&t, "%s/%s", main->wdpath, main->cnfname) < 0) {
            ifBLIN_QW0("No mem #1");
            ex = EX_OSERR;
            goto out;
        }
        free(main->cnfname);
        main->cnfname = t;
        ifBLIN_QX3("cnfname=%s~", main->cnfname);
    }
    if  (main->wdpath && main->pidname && (main->pidname[0] !=  '/')) {
        char *t;

        if  (asprintf(&t, "%s/%s", main->wdpath, main->pidname) < 0) {
            ifBLIN_QW0("No mem #2");
            ex = EX_OSERR;
            goto out;
        }
        free(main->pidname);
        main->pidname = t;
        ifBLIN_QX3("pidname=%s~\n", main->pidname);
    }
    if  (!(main->flags & POPUSH_DMN) && daemon(0, 0)) {
        ifBLIN_QW0("daemon");
        ex = EX_UNAVAILABLE;
        goto out;
    }
    mypid = getpid();
    if (emyspid) {
        if  (snprintf(myspid, 8, "%d", mypid) > 7) {
            ifBLIN_QW1("Too high pid %d", mypid);
        } else if (setenv(emyspid, myspid, 1) < 0) {
            ifBLIN_QW1("setenv %s=%s", emyspid, myspid);
    }   }
    if  ((main->kq = kqueue()) < 0) {
        ifBLIN_QW0("kqueue");
        ex = EX_OSERR;
        goto out;
    }
    popush_state = POPUSH_FSTRT;
    if  (signal(SIGHUP   , sighup) == SIG_ERR) ifBLIN_QW1("SIGHUP");
    if  (signal(SIGINT   , sigint) == SIG_ERR) ifBLIN_QW1("SIGINT");
    if  (signal(SIGTERM  , sigint) == SIG_ERR) ifBLIN_QW1("SIGTERM");
    if  (signal(SIGXCPU  , sigint) == SIG_ERR) ifBLIN_QW1("SIGXCPU");
    if  (signal(SIGXFSZ  , sigint) == SIG_ERR) ifBLIN_QW1("SIGXFSZ");
    if  (signal(SIGVTALRM, sigint) == SIG_ERR) ifBLIN_QW1("SIGVTALRM");
    for (; ;) {
        ifBLIN_QX4("Loop popush_state %08X\n", popush_state);
        if  ((popush_state & POPUSH_FSTOP) && (unlink(main->pidname) < 0)) {
            ifBLIN_QW1("PID file delete %s failed", main->pidname);
        }
        if  (popush_state & (POPUSH_FSTRT | POPUSH_FSIGN)) {
            if  ((pidfile = fopen(main->pidname, "w")) == NULL) {
                ifBLIN_QW1("PID file open %s failed", main->pidname);
            } else {
                if (fprintf(pidfile, "%d\n", mypid) < 0) {
                    ifBLIN_QW1("Can't write PID");
                }
                if  (fclose(pidfile)) {
                    ifBLIN_QW1("PID file not closed");
        }   }   }
        pidfile = NULL;
        if  (popush_state & (POPUSH_FSTRT | POPUSH_FSIGN | POPUSH_FSELF)) {
            if  (!(cfg = popush_syntan(main))) {
                ifBLIN_QW1("Syntax error");
            } else if (main->flags & POPUSH_ANA) {
                break;
            } else if (popush_diffmerge(main, cfg)) {
                /* ifBLIN_QW1("Configuration error")); */
            } else {
                popush_cnfree(main->cnf);
                main->cnf = cfg;
                free(main->actionbit);
                main->actionbit = NULL;
        }   }
        if  (!main->cnf) {
            ifBLIN_QX0("No config %s", main->cnfname);
            ex = EX_USAGE;
            goto out;
        }
        if  (  !main->actionbit
            && !(main->actionbit = calloc(((MULAR_NEXT(main->cnf->cluster)) + 7) >> 3, 1))
            ) {
            ifBLIN_QW0("create actionbit");
            ex = EX_TEMPFAIL;
            goto out;
        }
        if  (popush_state & POPUSH_FSTRT) popush_setact(main, main->cnf->strt);
        if  (popush_state & POPUSH_FSTOP) popush_setact(main, main->cnf->stop);
        if  (popush_state & POPUSH_FSIGN) popush_setact(main, main->cnf->sign);
        if  (popush_state & POPUSH_FSELF) popush_setaction(main, main->cnf->self);
        popush_state &= POPUSH_FSTOP;
        if  ((ex = popush_setevent(main))) {
            ifBLIN_QW0("setevent");
            goto out;
        }
        popush_execevent(main);
        if  (popush_state & POPUSH_FSTOP) {
            ifBLIN_QX0("killed");
            goto out;
        }
        if  (popush_state) continue;
        bzero(&evnt, sizeof(evnt));
        c = kevent(main->kq, NULL, 0, &evnt, 1, NULL);
        if  (c > 0) popush_setkact(main, &evnt);
    }
out:
    if  (!(popush_state & POPUSH_FSTOP)) {
        /* XXXX restart */
    }
    closelog();
    exit(ex);
#   undef blin_internal_flags
}
