/*-
 * Copyright (C)2005..2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2005..2019 @BABOLO http://www.babolo.ru/\n"
#ident "@(#) $Id: syntan.c,v 1.20 2021/08/28 13:43:26 babolo Exp $\n"

#define BLIN_COMPAT  4
#define MIFE_COMPAT  5
#define Bpars_COMPAT 4
#define MULAR_COMPAT 0

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <multilar.h>
#include <mife.h>
#include "popushd.h"
#include "event.h"
#include "sepa.h"

#ifndef BLIN_4STX24G
# define BLIN_4STX24G ((blin_statectl){{4, 0, 8, 24, ' ', 5, 1, 0, 0}})
#endif

static void
set(u_char *b, size_t n) {
    size_t i = n >> 3;
    u_char j = 1 << (n & 7);

    b[i] |= j;
}

static popush_cnf *
/*****************************************************************************************************
 *    popush_cnf                                                        *
 *****************************************************************************************************/
cnfin1(popush_main *main) {                                                                            /*
 *****************************************************************************************************/
#   define blin_internal_flags (main->flags & BLIN_MASK)
    size_t      u4[] = {1024, 512, 512, 512};
    size_t     *u8   = &(u4[1]);
    popush_cnf *cfg    = NULL;

    if  (!(cfg = calloc(1, sizeof(popush_cnf)))) {
        ifBLIN_QW0("No mem for popush_cnf");
        goto out;
    }
    if  (!(cfg->rawobj = mular_create(MULAR_ZERO, 3, 0, u4))) {
        ifBLIN_QW0("create rawobj");
        free(cfg);
        cfg = NULL;
        goto out;
    }
    cfg->rawobj->tofree = (babolo_tofree)babolo_freeparm;
    if  (!(cfg->reaction = mular_create(MULAR_ZERO, 3, 0, u4))) {
        ifBLIN_QW0("create reaction");
        mular_destroy(cfg->rawobj);
        free(cfg);
        cfg = NULL;
        goto out;
    }
    cfg->reaction->tofree = (babolo_tofree)babolo_freeparm;
    if  (!(cfg->cluster = mular_create(MULAR_ZERO, 3, sizeof(popush_clust), u8))) {
        ifBLIN_QW0("create cluster");
        mular_destroy(cfg->reaction);
        mular_destroy(cfg->rawobj);
        free(cfg);
        cfg = NULL;
        goto out;
    }
out:
    return(cfg);
#   undef blin_internal_flags
}

#define Ce  0x00800000 /*                */
#define Cw  0x00400000 /*   - */
#define Ca  0x00200000 /* &                 */
#define Cn  0x00100000 /*            */
#define Cf  0x00010000 /*               */
#define Cs  0x00008000 /*  SELF               */
#define Cb  0x00004000 /*               */
#define Co  0x00002000 /*                       */
#define Cp  0x00001000 /*                */
#define Cr  0x00000400 /*                      */
#define C_state  0xFF  /*           */

static const char pn[][3] = 
{ "- ", "- ", "- ", "- ", "- ", "- ", "- ", "- "
, "e ", "w ", "a ", "n ", "- ", "- ", "- ", "f "
, "s ", "b ", "o ", "p ", "z ", "r ", "- ", "- "
};

enum states {Sb, Sz, So, Sr, Sn, Sx};
/* Sb -   
   Sz -   
   So -  ()
   Sr -  ()
   Sn - 
   Sx -   ( :-)
 */
static const char statenames[][3] = {"Sb", "Sz", "So", "Sr", "Sn", "Sx"};

static const char clasnames[][3] =
{ "ER", "ZR", "AS", "ST", "SP", "SG", "SE", "FL"
, "NL", "_Z", "_A", "_T", "_P", "_G", "_L", "_F"
};

const u_int32_t popush_tin[POPUSH_MAXP][Sx] =
/*             Sb,               Sz,            So,         Sr,     Sn   */
{ { Ce|        Sb, Ce|           Sn, Ce|        Sn, Ce|     Sn, Cn| Sn} /* _EROR         */
, { Ce|        Sb, Ce|           Sz,            Sz,         Sz,     Sz} /* _ZERO         */
, { Ce|        Sb, Ce|           Sz, Ca|        So, Ca|     Sr, Cn| Sn} /* _ASYN         */
, {        Co| So,        Cb|Co| So,        Co| So, Ce|     Sn, Cn| Sn} /* _STRT         */
, {        Co| So,        Cb|Co| So,        Co| So, Ce|     Sn, Cn| Sn} /* _STOP         */
, {        Co| So,        Cb|Co| So,        Co| So, Ce|     Sn, Cn| Sn} /* _SIGN         */
, {     Cs|Co| So,     Cs|Cb|Co| So,     Cs|Co| So, Ce|     Sn, Cn| Sn} /* _SELF         */
, { Cf|    Cf| Sb, Cf|    Cf|    Sz, Cf|    Cf| So, Ce|     Sn, Cn| Sn} /* _FILE         */
, {            Sb,               Sz,     Cp|    Sz,         Sz,     Sz} /* _NULL         */
, { Ce|        Sb, Ce|           Sz,     Cp|Cr| Sr,     Cr| Sr, Cn| Sn} /* _ZERO | _FULL */
, { Ce|        Sb, Ce|           Sz,     Cp|Cr| Sr,     Cr| Sr, Cn| Sn} /* _ASYN | _FULL */
, { Cw|    Co| Sb, Cw|    Cb|Co| Sz, Cw|    Co| So, Ce|     Sn, Cn| Sn} /* _STRT | _FULL */
, { Cw|    Co| Sb, Cw|    Cb|Co| Sz, Cw|    Co| So, Ce|     Sn, Cn| Sn} /* _STOP | _FULL */
, { Cw|    Co| Sb, Cw|    Cb|Co| Sz, Cw|    Co| So, Ce|     Sn, Cn| Sn} /* _SIGN | _FULL */
, { Cw| Cs|Co| Sb, Cw| Cs|Cb|Co| Sz, Cw| Cs|Co| So, Ce|     Sn, Cn| Sn} /* _SELF | _FULL */
, {        Co| So,        Cb|Co| So,        Co| So, Ce|     Sn, Cn| Sn} /* _FILE | _FULL */
};
/* Cp  [POPUSH_NULL][So]      .
 *       ,
 *        ask   .
 * XXXX
 */

static int
run1(popush_main *main, popush_cnf *cfg) {
#   define blin_internal_flags (main->flags & BLIN_MASK)
    int           ex = EX_OK;
    u_int32_t     control;
    u_int32_t     nextat;
    u_char       *nextrc;
    ssize_t       rcsize; /*     mife   */
    u_int32_t     state;
    int           line;
    off_t         off;
    babolo_parm **t;
    babolo_parm  *s;
    popush_clust *p;

    ifBLIN_QX3("+");
    if  (main->cnf) {
        cfg->konfik.evnt.ident = main->cnf->konfik.evnt.ident;
        lseek(cfg->konfik.evnt.ident, 0, SEEK_SET);
    } else {
        int y = open(main->cnfname, O_RDONLY);

        if  (0 > y) {
            ifBLIN_QW0("config file open %s failed", main->cnfname);
            ex = EX_NOINPUT;
            goto out;
        }
        cfg->konfik.evnt.ident = y;
    }
    if  (!(main->cnfile = mife_init(MIFE_FULL))) {
        ifBLIN_QW0("mife_init failed");
        ex = EX_NOINPUT;
        goto out;
    }
    if  (0 > mife_ctlfdsc(main->cnfile, cfg->konfik.evnt.ident)) {
        ifBLIN_QW0("config file mife open %s failed", main->cnfname);
        ex = EX_NOINPUT;
        goto out;
    }
    if  (0 > mife_ctleoli(main->cnfile, "\n")) {
        ifBLIN_QW0("mife_eolinit");
        ex = EX_SOFTWARE;
        goto out;
    }
    ifBLIN_QO4 blin_stateheader(BLIN_4STX24G, pn);
    for ( line = 1, state = Sb, nextrc = NULL
        ; state < Sx
        ; state = control & C_state
        ) {
        if  (!!nextrc) {
            off = main->cnfile->offset + (nextrc - (u_char*)main->cnfile->buffer);
        } else {
            off = 0;
        }
        if  (0 > (rcsize = mife_read(main->cnfile, 0, off))) {
            ifBLIN_QW0("config file read failed");
            ex = EX_IOERR;
            goto out;
        } else if (rcsize == 0) {
            break;
        }
        if  (!(nextrc = mife_get(main->cnfile, off))) {
            ifBLIN_QW0("mife_get");
            ex = EX_SOFTWARE;
            goto out;
        }
        nextat = 0;
        if  (!(s = babolo_getparm(Bpars_FIXB | Bpars_XXLN, (char **)&nextrc, "#", main->absq))) {
            ifBLIN_QW0("babolo_getparm error in line %d", line);
            ex = EX_SOFTWARE;
            goto out;
        } else if (s->argc == 0) {
            nextat = POPUSH_NULL;
        } else {
            nextat = babolo_testword(&sepa, (s->argv)[0]);
        }
        line += s->cntr;
        if  ((nextat != POPUSH_EROR) && (s->argc > 1)) nextat |= POPUSH_FULL;
        control = popush_tin[nextat][state];
        ifBLIN_QO4 blin_statebody( BLIN_4STX24G
                                 , pn
                                 , statenames
                                 , clasnames[nextat]
                                 , (char*)nextrc
                                 , line
                                 , control
                                 , state
                                 , s->argc
                                 , nextat
                                 );
        if  (control & Ce) {
            ifBLIN_QX1("Config error in line %d", line - 1);
        }
        if  (control & Cf) {
            ifBLIN_QX1("File without events in line %d", line);
            babolo_freeparm(s);
        }
        if  (control & Ca) {
            ifBLIN_QX1("Lonely & in line %d", line);
        }
        if  (control & Cn) {
            ifBLIN_QX1("Skip line %d for previous error(s)", line);
        }
        if  (control & Cf) {
            ifBLIN_QX1("File without events in line %d", line);
            babolo_freeparm(s);
        }
        if  (control & Cs) {
            ifBLIN_QX4("SELF %d", line);
            cfg->self = 3;
        }
        if  (control & Cb) {
            ifBLIN_QX4( "New obj=%s~ rawobj[%"BLIN_U"]=NULL"
                      , (s->argv)[0]
                      , MULAR_NEXT(cfg->rawobj)
                      );
            if  (!(t = mular_add(cfg->rawobj))) {
                ifBLIN_QW0("mular_add rawobj");
                ex = EX_SOFTWARE;
                goto out;
            }
            *t = NULL;
        }
        if  (control & Co) {
            ifBLIN_QX4( "Obj=%s~ rawobj[%"BLIN_U"]=%"BLIN_X
                      , s->argv[0]
                      , MULAR_NEXT(cfg->rawobj)
                      , BLIN_I(s)
                      );
            if  (!(t = mular_add(cfg->rawobj))) {
                ifBLIN_QW0("mular_add rawobj");
                ex = EX_SOFTWARE;
                goto out;
            }
            *t = s;
        }
        if  (control & Cp) {
            ifBLIN_QX4("cluster[%"BLIN_U"]=%"BLIN_U
                      , MULAR_NEXT(cfg->cluster)
                      , MULAR_NEXT(cfg->reaction)
                      );
            if  (!(p = mular_add(cfg->cluster))) {
                ifBLIN_QW0("mular_add cluster");
                ex = EX_SOFTWARE;
                goto out;
            }
            p->clust = MULAR_NEXT(cfg->reaction);
        }
        if  (control & Cr) {
            ifBLIN_QX4("React=%s~ reaction[%"BLIN_U"]=%"BLIN_X
                      , s->argv[1]
                      , MULAR_NEXT(cfg->reaction)
                      , BLIN_I(s)
                      );
            if  (!(t = mular_add(cfg->reaction))) {
                ifBLIN_QW0("mular_add reaction");
                ex = EX_SOFTWARE;
                goto out;
            }
            *t = s;
    }   }
out:
    if  (mife_fini(main->cnfile)) {
        ifBLIN_QW1("config not closed");
    }
    main->cnfile = 0;
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
run2(popush_main *main, popush_cnf *cfg) {
#   define blin_internal_flags (main->flags & BLIN_MASK)
    int     ex   = EX_OK;
    size_t  l    = MULAR_NEXT(cfg->cluster);
    size_t  ll   = (l + 7) / 8;
    size_t  ux[] = {0, 512, 512};
    size_t  objn;
    size_t  objp;
    u_char *a;

    ifBLIN_QX3("+ %"BLIN_U" %"BLIN_U, l, ll);
    if  (ll == 0) {
        ifBLIN_QX0("empty input file");
        ex = EX_DATAERR;
        goto out;
    }
    ux[0] = 4096 / ll;
    if  (ux[0] < 12) ux[0] = 12;
    ifBLIN_QX4("ask=%"BLIN_U, ux[0]);
    if  (!(cfg->ask = mular_create(MULAR_ZERO, 3, ll, ux))) {
        ifBLIN_QW0("create ask %u(%u)", ll, l);
        ex = EX_SOFTWARE;
        goto out;
    }
    cfg->strt = 0;
    if  (cfg->strt != MULAR_NEXT(cfg->ask)) {
        ifBLIN_QX0("strt %u <> %u", cfg->strt, MULAR_NEXT(cfg->ask));
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (!(a = mular_add(cfg->ask))) {
        ifBLIN_QW0("mular_add ask strt");
        ex = EX_SOFTWARE;
        goto out;
    }
    cfg->stop = 1;
    if  (cfg->stop != MULAR_NEXT(cfg->ask)) {
        ifBLIN_QX0("stop %u <> %u", cfg->stop, MULAR_NEXT(cfg->ask));
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (!(a = mular_add(cfg->ask))) {
        ifBLIN_QW0("mular_add ask stop");
        ex = EX_SOFTWARE;
        goto out;
    }
    cfg->sign = 2;
    if  (cfg->sign != MULAR_NEXT(cfg->ask)) {
        ifBLIN_QX0("sign %u <> %u", cfg->sign, MULAR_NEXT(cfg->ask));
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (!(a = mular_add(cfg->ask))) {
        ifBLIN_QW0("mular_add ask sign");
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (cfg->self) {
        if  (cfg->self != MULAR_NEXT(cfg->ask)) {
            ifBLIN_QX0("self %u <> %u", cfg->self, MULAR_NEXT(cfg->ask));
            ex = EX_SOFTWARE;
            goto out;
        }
        if  (!(a = mular_add(cfg->ask))) {
            ifBLIN_QW0("mular_add ask self");
            ex = EX_SOFTWARE;
            goto out;
    }   }
    ux[0] = 8192 / sizeof(popush_obj);
    ifBLIN_QX4("object=%"BLIN_U, ux[0]);
    if  (!(cfg->object = mular_create(MULAR_ZERO, 3, sizeof(popush_obj), ux))) {
        ifBLIN_QW0("create object");
        ex = EX_SOFTWARE;
        goto out;
    }
    ifBLIN_QX4( "rawobj max %"BLIN_U", cluster max %"BLIN_U
              , MULAR_NEXT(cfg->rawobj), MULAR_NEXT(cfg->cluster)
              );
    for (objn = 0, objp = 0; !ex && objp < MULAR_NEXT(cfg->rawobj); objp++) {
        size_t       c0 = 0;
        size_t       c1 = 0;
        size_t      *g0;
        size_t      *g1;
        babolo_parm *p;
        popush_obj  *o;
        size_t      *v;
        size_t       n;
        int          i;
        u_int32_t    j;
        u_int32_t    t;

        p = *(babolo_parm**)(mular_getix(cfg->rawobj, objp));
        t = 0;
        if  (objn < MULAR_NEXT(cfg->cluster)) {
            t |= 2;
            g0 = mular_getix(cfg->cluster, objn);
            if  (!g0) {
                /* XXXX error */
            }
            c0 = *g0;
            if  (c0 == MULAR_NEXT(cfg->object)) t |= 8;
        }
        if  ((objn + 1) < MULAR_NEXT(cfg->cluster)) {
            t |= 1;
            g1 = mular_getix(cfg->cluster, objn + 1);
            if  (!g1) {
                /* XXXX error */
            }
            c1 = *g1;
            if  (c0 == c1) t |= 16;
        } else if ((objn + 1) == MULAR_NEXT(cfg->cluster)) t |= 4;
/****************
   16 8 4 2 1
    0 0 0 0 0 ERROR
    0 0 0 1 1
    0 0 1 1 0
    0 1 0 1 1
    0 1 1 1 0
    1 0 0 1 1 NEXT
    1 1 0 1 1 NEXT
 ****************/
        ifBLIN_QO4 {
            if  (t & 1) {
                ifBLIN_QX4( "%02X: clu[%"BLIN_U"]=%"BLIN_U" <=> clu[%"BLIN_U"]=%"BLIN_U
                            ", raw[%"BLIN_U"]"
                          , t
                          , objn
                          , c0
                          , objn + 1
                          , c1
                          , objp
                          );
            } else if (t & 2) {
                ifBLIN_QX4( "%02X: clu[%"BLIN_U"]=%"BLIN_U" <=> clu[%"BLIN_U"] NO, raw[%"BLIN_U"]"
                          , t
                          , objn
                          , c0
                          , objn + 1
                          , objp
                          );
            } else {
                ifBLIN_QX4( "%02X: clu[%"BLIN_U"] NO <=> clu[%"BLIN_U"] NO, raw[%"BLIN_U"]"
                          , t
                          , objn
                          , objn + 1
                          , objp
                          );
        }   }
        if  (!t) { /* ERROR */
            ex = EX_SOFTWARE;
            ifBLIN_QX0("Internal parser error");
            break;
        } else if (!p) {
            objn++;
            ifBLIN_QX4(" NO");
        } else if (t & 16) { /* NEXT */
            ifBLIN_QX4(" ==");
        } else {
        ;   v = NULL;
        ;   t = babolo_testword(&sepa, p->argv[0]);
        ;   ifBLIN_QX4("=%02X %s~...", t, p->argv[0]);
        ;   switch (t) {
            case POPUSH_STRT: ifBLIN_QX4("STRT"); v = &(cfg->strt); break;
        ;   case POPUSH_STOP: ifBLIN_QX4("STOP"); v = &(cfg->stop); break;
        ;   case POPUSH_SIGN: ifBLIN_QX4("SIGN"); v = &(cfg->sign); break;
        ;   case POPUSH_SELF:
        ;       ifBLIN_QX4("SELF");
        ;       if  (cfg->self) v = &(cfg->self);
        ;       break;
        ;   case POPUSH_FILE:
        ;       ifBLIN_QX4("FILE");
        ;       for (n = 0, o = NULL; n < MULAR_NEXT(cfg->object); n++) {
        ;           if  (!(o = mular_getix(cfg->object, n))) {
        ;               ifBLIN_QW1("mular_getix object %"BLIN_U, n);
        ;           } else if (strcmp((char *)p->argv[0], o->fname) == 0) break;
        ;           o = NULL;
        ;       }
        ;       if  (!o) {
        ;           /*      ,    ,
        ;            *       
        ;            *  popush_diffmerge().  XXXX
        ;            */
        ;           ifBLIN_QX4(" %"BLIN_U" %s", MULAR_NEXT(cfg->object), p->argv[0]);
        ;           if  (!(o = mular_add(cfg->object))) {
        ;               ifBLIN_QW0("add object");
        ;               ex = EX_SOFTWARE;
        ;               goto out;
        ;           }
        ;           o->fname = (char *)p->argv[0];
        ;           o->evnt.filter = EVFILT_VNODE;
        ;           o->evnt.flags = 0;
        ;       }
        ;       for (i = 1; i < p->argc; i++) {
        ;           switch ((j = babolo_testword(&event, p->argv[i]))) {
                    case NOTE_DELETE:
        ;               ifBLIN_QX4("DELETE");
        ;               v = &(o->fdel);
        ;               break;
        ;           case NOTE_WRITE :
        ;               ifBLIN_QX4("WRITE");
        ;               v = &(o->fwrt);
        ;               break;
        ;           case NOTE_EXTEND:
        ;               ifBLIN_QX4("EXTEND\n");
        ;               v = &(o->fext);
        ;               break;
        ;           case NOTE_ATTRIB:
        ;               ifBLIN_QX4("ATTRIB");
        ;               v = &(o->fatr);
        ;               break;
        ;           case NOTE_LINK  :
        ;               ifBLIN_QX4("LINK");
        ;               v = &(o->flnk);
        ;               break;
        ;           case NOTE_RENAME:
        ;               ifBLIN_QX4("RENAME");
        ;               v = &(o->fren);
        ;               break;
        ;           case NOTE_REVOKE:
        ;               ifBLIN_QX4("REVOKE");
        ;               v = &(o->frev);
        ;               break;
        ;           default:
        ;               ifBLIN_QX1("rawobj #2");
        ;               v = NULL;
        ;           }
        ;           if  (v) {
        ;               ifBLIN_QX4(" %"BLIN_U" %08X", MULAR_NEXT(cfg->ask), j);
        ;               o->evnt.fflags |= j;
        ;               *v = MULAR_NEXT(cfg->ask);
        ;               if  (!mular_add(cfg->ask)) {
        ;                   ifBLIN_QX1("add ask");
        ;               }
        ;               set(/* XXXX */mular_getix(cfg->ask, *v), objn);
        ;               v = NULL;
        ;       }   }
        ;       break;
        ;   default:
        ;       ifBLIN_QX1("rawobj #1");
        ;       ex = EX_SOFTWARE;
        ;       goto out;
        ;   }
        ;   if  (v) {
        ;       ifBLIN_QX3(" %"BLIN_U, MULAR_NEXT(cfg->ask));
        ;       set(/* XXXX */mular_getix(cfg->ask, *v), objn);
        ;   }
    }   }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static void
dumpr1(popush_cnf *cfg) {
    babolo_parm **t;
    popush_clust *c;
    size_t        n;
    size_t        i;

    fprintf(stderr, "+################################ run1 dump\n## rawobj\n");
    for (n = 0; n < MULAR_NEXT(cfg->rawobj); n++) {
        t = mular_getix(cfg->rawobj, n);
        if  (!t) {
            fprintf(stderr, "## raw[%"BLIN_D"] NO\n", n);
        } else if (!*t) {
            fprintf(stderr, "## raw[%"BLIN_D"] NULL\n", n);
        } else {
            fprintf( stderr, "## raw[%"BLIN_D"] %08X %"BLIN_D" %d\n"
                   , n, (*t)->flags, (*t)->cntr, (*t)->argc
                   );
            for (i = 0; i < (*t)->argc; i++) {
                if  ((*t)->argv[i]) {
                    fprintf(stderr, " [%"BLIN_D"]=%s~\n", i, (*t)->argv[i]);
                } else {
                    fprintf(stderr, " [%"BLIN_D"] NULL\n", i);
    }   }   }   }
    fprintf(stderr, "## reaction\n");
    for (n = 0; n < MULAR_NEXT(cfg->reaction); n++) {
        t = mular_getix(cfg->reaction, n);
        if  (!t) {
            fprintf(stderr, "## rea[%"BLIN_D"] NO\n", n);
        } else if (!*t) {
            fprintf(stderr, "## rea[%"BLIN_D"] NULL\n", n);
        } else {
            fprintf( stderr, "## rea[%"BLIN_D"] %08X %"BLIN_D" %d\n"
                   , n, (*t)->flags, (*t)->cntr, (*t)->argc
                   );
            for (i = 0; i < (*t)->argc; i++) {
                if  ((*t)->argv[i]) {
                    fprintf(stderr, " [%"BLIN_D"]=%s~\n", i, (*t)->argv[i]);
                } else {
                    fprintf(stderr, " [%"BLIN_D"] NULL\n", i);
    }   }   }   }
    fprintf(stderr, "## cluster\n");
    for (n = 0; n < MULAR_NEXT(cfg->cluster); n++) {
        c = mular_getix(cfg->cluster, n);
        if  (c) {
            fprintf(stderr, "clu[%"BLIN_D"]=%"BLIN_D"(rea)\n", n, c->clust);
        } else {
            fprintf(stderr, "clu[%"BLIN_D"] NULL\n", n);
    }   }
    fprintf(stderr, "-################################ run1 dump\n");
}

static void
dubit(u_char *b, size_t l, char *c) {
    size_t k = 0;

    for (size_t n = 0; n < l; n++) {
        size_t i = n >> 3;
        u_char j = 1 << (n & 7);

        if  (!k) {
            fprintf(stderr, "%s", c);
            k++;
        }
        if  (b[i] & j) fprintf(stderr, ",%"BLIN_U"", n);
    }
    if  (k) fprintf(stderr, "\n");
}

static void
dumpcf(popush_cnf *cfg) {
    babolo_parm **p;
    popush_obj   *o;
    size_t        n;
    size_t        m;
    size_t        i;

    fprintf(stderr, "+################################ config dump\n");
    dubit(mular_getix(cfg->ask, cfg->strt), MULAR_NEXT(cfg->cluster), "strt");
    dubit(mular_getix(cfg->ask, cfg->stop), MULAR_NEXT(cfg->cluster), "stop");
    dubit(mular_getix(cfg->ask, cfg->sign), MULAR_NEXT(cfg->cluster), "sign");
    if  (cfg->self) dubit(mular_getix(cfg->ask, cfg->self), MULAR_NEXT(cfg->cluster), "self");
    for (n = 0; n < MULAR_NEXT(cfg->object); n++) {
        if  (!(o = mular_getix(cfg->object, n))) {
            fprintf(stderr, "NULL object %"BLIN_U"\n", n);
            continue;
        }
        fprintf(stderr, "%s\n", o->fname);
        if  (o->fdel) dubit(mular_getix(cfg->ask, o->fdel), MULAR_NEXT(cfg->cluster), " fdel");
        if  (o->fwrt) dubit(mular_getix(cfg->ask, o->fwrt), MULAR_NEXT(cfg->cluster), " fwrt");
        if  (o->fext) dubit(mular_getix(cfg->ask, o->fext), MULAR_NEXT(cfg->cluster), " fext");
        if  (o->fatr) dubit(mular_getix(cfg->ask, o->fatr), MULAR_NEXT(cfg->cluster), " fatr");
        if  (o->flnk) dubit(mular_getix(cfg->ask, o->flnk), MULAR_NEXT(cfg->cluster), " flnk");
        if  (o->fren) dubit(mular_getix(cfg->ask, o->fren), MULAR_NEXT(cfg->cluster), " fren");
        if  (o->frev) dubit(mular_getix(cfg->ask, o->frev), MULAR_NEXT(cfg->cluster), " frev");
    }
    for (m = n = 0; n < MULAR_NEXT(cfg->reaction); n++) {
        int k;
        p = mular_getix(cfg->reaction, n);
        k = 0;
        while ((m < MULAR_NEXT(cfg->cluster)) && (n == (*((size_t*)mular_getix(cfg->cluster, m))))) {
            k = 1;
            m++;
        }
        if  (k) fprintf(stderr, "## rea[%"BLIN_U"] for clu[%"BLIN_U"]\n", n, m - 1);
        if  (*p) {
            for (i = 0; i < (*p)->argc; i++) {
                fprintf(stderr, " %s", (*p)->argv[i]);
            }
        } else fprintf(stderr, "NULL");
        fprintf(stderr, "\n");
    }
    fprintf(stderr, "-################################ config dump\n");
}

popush_cnf *
popush_syntan(popush_main *main) {
#   define blin_internal_flags (main->flags & BLIN_MASK)
    int         ex = EX_OK;
    popush_cnf *cfg;

    ifBLIN_QX3("+");
    if  (!(cfg = cnfin1(main))) {
        ifBLIN_QW0("cnfin1");
        ex = EX_UNAVAILABLE;
        goto out;
    }
    if  ((ex = run1(main, cfg))) {
        ifBLIN_QW0("run1");
        ex = EX_UNAVAILABLE;
        goto out;
    }
    if  (main->flags & POPUSH_ANA) dumpr1(cfg);
    if  ((ex = run2(main, cfg))) goto out;
    if  (main->flags & POPUSH_ANA) dumpcf(cfg);
out:if  (ex) {
        popush_cnfree(cfg);
        cfg = NULL;
    }
    ifBLIN_QX3("- %d %"BLIN_X, ex, BLIN_I(cfg));
    return(cfg);
#   undef blin_internal_flags
}
